@@ -78,6 +78,7 @@ struct a2dp_setup_cb {
 };
 
 struct a2dp_setup {
+	struct audio_device *dev;
 	struct avdtp *session;
 	struct a2dp_sep *sep;
 	struct avdtp_stream *stream;
@@ -136,11 +137,11 @@ static void setup_unref(struct a2dp_setup *setup)
 
 static struct audio_device *a2dp_get_dev(struct avdtp *session)
 {
-	bdaddr_t addr;
+	bdaddr_t src, dst;
 
-	avdtp_get_peers(session, NULL, &addr);
+	avdtp_get_peers(session, &src, &dst);
 
-	return manager_find_device(&addr, NULL, FALSE);
+	return manager_find_device(NULL, &src, &dst, NULL, FALSE);
 }
 
 static gboolean finalize_config(struct a2dp_setup *s)
@@ -183,7 +184,7 @@ static gboolean finalize_resume(struct a2dp_setup *s)
 	for (l = s->cb; l != NULL; l = l->next) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb->resume_cb) {
+		if (cb && cb->resume_cb) {
 			cb->resume_cb(s->session, s->err, cb->user_data);
 			cb->resume_cb = NULL;
 			setup_unref(s);
@@ -243,9 +244,8 @@ static struct a2dp_setup *find_setup_by_dev(struct audio_device *dev)
 
 	for (l = setups; l != NULL; l = l->next) {
 		struct a2dp_setup *setup = l->data;
-		struct audio_device *setup_dev = a2dp_get_dev(setup->session);
 
-		if (setup_dev == dev)
+		if (setup->dev == dev)
 			return setup;
 	}
 
@@ -507,8 +507,7 @@ static void setconf_cfm(struct avdtp *session, struct avdtp_local_sep *sep,
 
 	ret = avdtp_open(session, stream);
 	if (ret < 0) {
-		error("Error on avdtp_open %s (%d)", strerror(-ret),
-				-ret);
+		error("Error on avdtp_open %s (%d)", strerror(-ret), -ret);
 		setup->stream = NULL;
 		finalize_config_errno(setup, ret);
 	}
@@ -520,9 +519,9 @@ static gboolean getconf_ind(struct avdtp *session, struct avdtp_local_sep *sep,
 	struct a2dp_sep *a2dp_sep = user_data;
 
 	if (a2dp_sep->type == AVDTP_SEP_TYPE_SINK)
-		debug("Sink %p: Get_Configuration_Ind");
+		debug("Sink %p: Get_Configuration_Ind", sep);
 	else
-		debug("Source %p: Get_Configuration_Ind");
+		debug("Source %p: Get_Configuration_Ind", sep);
 	return TRUE;
 }
 
@@ -759,25 +758,34 @@ static gboolean a2dp_reconfigure(gpointer data)
 		break;
 	}
 
+	if (!codec_cap) {
+		error("Cannot find capabilities to reconfigure");
+		posix_err = -EINVAL;
+		goto failed;
+	}
+
 	posix_err = avdtp_get_seps(setup->session, AVDTP_SEP_TYPE_SINK,
 					codec_cap->media_type,
 					codec_cap->media_codec_type,
 					&lsep, &rsep);
 	if (posix_err < 0) {
 		error("No matching ACP and INT SEPs found");
-		finalize_config_errno(setup, posix_err);
+		goto failed;
 	}
 
 	posix_err = avdtp_set_configuration(setup->session, rsep, lsep,
 						setup->client_caps,
 						&setup->stream);
 	if (posix_err < 0) {
-		error("avdtp_set_configuration: %s",
-			strerror(-posix_err));
-		finalize_config_errno(setup, posix_err);
+		error("avdtp_set_configuration: %s", strerror(-posix_err));
+		goto failed;
 	}
 
 	return FALSE;
+
+failed:
+	finalize_config_errno(setup, posix_err);
+	return FALSE;
 }
 
 static void close_cfm(struct avdtp *session, struct avdtp_local_sep *sep,
@@ -1189,6 +1197,8 @@ void a2dp_unregister(const bdaddr_t *src)
 	g_slist_foreach(server->sources, (GFunc) a2dp_unregister_sep, NULL);
 	g_slist_free(server->sources);
 
+	avdtp_exit(src);
+
 	if (server->source_record_id)
 		remove_record_from_server(server->source_record_id);
 
@@ -1211,6 +1221,8 @@ gboolean a2dp_source_cancel(struct audio_device *dev, unsigned int id)
 	struct a2dp_setup *setup;
 	GSList *l;
 
+	debug("a2dp_source_cancel()");
+
 	setup = find_setup_by_dev(dev);
 	if (!setup)
 		return FALSE;
@@ -1225,7 +1237,7 @@ gboolean a2dp_source_cancel(struct audio_device *dev, unsigned int id)
 	}
 
 	if (!cb_data)
-		return FALSE;
+		error("a2dp_source_cancel: no matching callback with id %u", id);
 
 	setup->cb = g_slist_remove(setup->cb, cb_data);
 	g_free(cb_data);
@@ -1238,14 +1250,48 @@ gboolean a2dp_source_cancel(struct audio_device *dev, unsigned int id)
 	return TRUE;
 }
 
-unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
-				GSList *caps, void *user_data)
+struct a2dp_sep *a2dp_source_get(struct avdtp *session,
+				struct avdtp_remote_sep *rsep)
+{
+	GSList *l;
+	struct a2dp_server *server;
+	struct avdtp_service_capability *cap;
+	struct avdtp_media_codec_capability *codec_cap = NULL;
+	bdaddr_t src;
+
+	avdtp_get_peers(session, &src, NULL);
+	server = find_server(servers, &src);
+	if (!server)
+		return NULL;
+
+	cap = avdtp_get_codec(rsep);
+	codec_cap = (void *) cap->data;
+
+	for (l = server->sources; l != NULL; l = l->next) {
+		struct a2dp_sep *sep = l->data;
+
+		if (sep->locked)
+			continue;
+
+		if (sep->codec != codec_cap->media_codec_type)
+			continue;
+
+		if (!sep->stream || avdtp_has_stream(session, sep->stream))
+			return sep;
+	}
+
+	return NULL;
+}
+
+unsigned int a2dp_source_config(struct avdtp *session, struct a2dp_sep *sep,
+				a2dp_config_cb_t cb, GSList *caps,
+				void *user_data)
 {
 	struct a2dp_setup_cb *cb_data;
 	GSList *l;
 	struct a2dp_server *server;
 	struct a2dp_setup *setup;
-	struct a2dp_sep *sep = NULL, *tmp;
+	struct a2dp_sep *tmp;
 	struct avdtp_local_sep *lsep;
 	struct avdtp_remote_sep *rsep;
 	struct avdtp_service_capability *cap;
@@ -1271,25 +1317,8 @@ unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
 	if (!codec_cap)
 		return 0;
 
-	for (l = server->sources; l != NULL; l = l->next) {
-		tmp = l->data;
-
-		if (tmp->locked)
-			continue;
-
-		if (tmp->codec != codec_cap->media_codec_type)
-			continue;
-
-		if (!tmp->stream || avdtp_has_stream(session, tmp->stream)) {
-			sep = tmp;
-			break;
-		}
-	}
-
-	if (!sep) {
-		error("a2dp_source_cfg: no available SEP found");
+	if (sep->codec != codec_cap->media_codec_type)
 		return 0;
-	}
 
 	debug("a2dp_source_config: selected SEP %p", sep->sep);
 
@@ -1302,6 +1331,7 @@ unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
 	if (!setup) {
 		setup = g_new0(struct a2dp_setup, 1);
 		setup->session = avdtp_ref(session);
+		setup->dev = a2dp_get_dev(session);
 		setups = g_slist_append(setups, setup);
 	}
 
@@ -1347,9 +1377,10 @@ unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
 		break;
 	case AVDTP_STATE_OPEN:
 	case AVDTP_STATE_STREAMING:
-		if (avdtp_stream_has_capabilities(setup->stream, caps))
+		if (avdtp_stream_has_capabilities(setup->stream, caps)) {
+			debug("Configuration match: resuming");
 			g_idle_add((GSourceFunc) finalize_config, setup);
-		else if (!setup->reconfigure) {
+		} else if (!setup->reconfigure) {
 			setup->reconfigure = TRUE;
 			if (avdtp_close(session, sep->stream) < 0) {
 				error("avdtp_close failed");
@@ -1385,6 +1416,7 @@ unsigned int a2dp_source_resume(struct avdtp *session, struct a2dp_sep *sep,
 	if (!setup) {
 		setup = g_new0(struct a2dp_setup, 1);
 		setup->session = avdtp_ref(session);
+		setup->dev = a2dp_get_dev(session);
 		setups = g_slist_append(setups, setup);
 	}
 
@@ -1416,7 +1448,7 @@ unsigned int a2dp_source_resume(struct avdtp *session, struct a2dp_sep *sep,
 			g_idle_add((GSourceFunc) finalize_resume, setup);
 		break;
 	default:
-		error("SEP in bad state");
+		error("SEP in bad state for resume");
 		goto failed;
 	}
 
@@ -1443,6 +1475,7 @@ unsigned int a2dp_source_suspend(struct avdtp *session, struct a2dp_sep *sep,
 	if (!setup) {
 		setup = g_new0(struct a2dp_setup, 1);
 		setup->session = avdtp_ref(session);
+		setup->dev = a2dp_get_dev(session);
 		setups = g_slist_append(setups, setup);
 	}
 
@@ -1466,7 +1499,7 @@ unsigned int a2dp_source_suspend(struct avdtp *session, struct a2dp_sep *sep,
 		}
 		break;
 	default:
-		error("SEP in bad state for resume");
+		error("SEP in bad state for suspend");
 		goto failed;
 	}
 
@@ -1517,3 +1550,45 @@ gboolean a2dp_sep_unlock(struct a2dp_sep *sep, struct avdtp *session)
 	return TRUE;
 }
 
+gboolean a2dp_sep_get_lock(struct a2dp_sep *sep)
+{
+	return sep->locked;
+}
+
+static int stream_cmp(gconstpointer data, gconstpointer user_data)
+{
+	const struct a2dp_sep *sep = data;
+	const struct avdtp_stream *stream = user_data;
+
+	return (sep->stream != stream);
+}
+
+struct a2dp_sep *a2dp_get_sep(struct avdtp *session,
+				struct avdtp_stream *stream)
+{
+	struct a2dp_server *server;
+	bdaddr_t src, dst;
+	GSList *l;
+
+	avdtp_get_peers(session, &src, &dst);
+
+	for (l = servers; l; l = l->next) {
+		server = l->data;
+
+		if (bacmp(&src, &server->src) == 0)
+			break;
+	}
+
+	if (!l)
+		return NULL;
+
+	l = g_slist_find_custom(server->sources, stream, stream_cmp);
+	if (l)
+		return l->data;
+
+	l = g_slist_find_custom(server->sinks, stream, stream_cmp);
+	if (l)
+		return l->data;
+
+	return NULL;
+}
